/*

Audio mixer information utility.

Written by Eugene V. Muzychenko.

Build command line:

cl mixerinfo.cpp

*/



#pragma optimize ("gt", on)

#if _MSC_VER > 1200

  #pragma runtime_checks ("", off)

#endif

#pragma warning (disable: 4115 4127 4201 4214 4514)


#ifdef _DEBUG

  #pragma warning (disable: 4505)

#endif


#define STRICT

#define WIN32_LEAN_AND_MEAN 1

#include <windows.h>
#include <mmsystem.h>
#include <shlwapi.h>

#include <stdio.h>
#include <stdlib.h>





// #########################################################################
//
// Defines
// =======
// 
// #########################################################################





#define strlen lstrlen
#define strcpy lstrcpy
#define strcat lstrcat





// #########################################################################
//
// Linker parameters
// =================
// 
// #########################################################################





#pragma comment (lib, "kernel32.lib")
#pragma comment (lib, "user32.lib")
#pragma comment (lib, "shlwapi.lib")
#pragma comment (lib, "winmm.lib")

#pragma comment (linker, "/subsystem:console")
#pragma comment (linker, "/entry:WinEntry")

#if _MSC_VER <= 1300

  #pragma comment (linker, "/opt:nowin98")

#endif





// #########################################################################
//
// Defines
// =======
// 
// #########################################################################





#define Unused (void)





// #########################################################################
//
// Structures/classes
// ==================
// 
// #########################################################################





struct ValueNameEntry {

  PCSTR Name;
  DWORD Value;

};





class String {

  PSTR Str;

  public:

  String (PCSTR S) {
    Str = strcpy (PSTR (malloc (strlen (S) + 1)), S);
  }

  ~String (void) { free (Str); }

  operator PCSTR (void) { return Str; }

};





// #########################################################################
//
// Constants
// =========
// 
// #########################################################################





ValueNameEntry const LineStatusFlags [] = {

  {"Source", MIXERLINE_LINEF_SOURCE},
  {"Active", MIXERLINE_LINEF_ACTIVE},
  {"Disconnected", MIXERLINE_LINEF_DISCONNECTED},

  {NULL}

};





ValueNameEntry const CompTypes [] = {

  {"DST_DIGITAL", MIXERLINE_COMPONENTTYPE_DST_DIGITAL},
  {"DST_HEADPHONES", MIXERLINE_COMPONENTTYPE_DST_HEADPHONES},
  {"DST_LINE", MIXERLINE_COMPONENTTYPE_DST_LINE},
  {"DST_MONITOR", MIXERLINE_COMPONENTTYPE_DST_MONITOR},
  {"DST_SPEAKERS", MIXERLINE_COMPONENTTYPE_DST_SPEAKERS},
  {"DST_TELEPHONE", MIXERLINE_COMPONENTTYPE_DST_TELEPHONE},
  {"DST_UNDEFINED", MIXERLINE_COMPONENTTYPE_DST_UNDEFINED},
  {"DST_VOICEIN", MIXERLINE_COMPONENTTYPE_DST_VOICEIN},
  {"DST_WAVEIN", MIXERLINE_COMPONENTTYPE_DST_WAVEIN},
  {"SRC_ANALOG", MIXERLINE_COMPONENTTYPE_SRC_ANALOG},
  {"SRC_AUXILIARY", MIXERLINE_COMPONENTTYPE_SRC_AUXILIARY},
  {"SRC_COMPACTDISC", MIXERLINE_COMPONENTTYPE_SRC_COMPACTDISC},
  {"SRC_DIGITAL", MIXERLINE_COMPONENTTYPE_SRC_DIGITAL},
  {"SRC_LINE", MIXERLINE_COMPONENTTYPE_SRC_LINE},
  {"SRC_MICROPHONE", MIXERLINE_COMPONENTTYPE_SRC_MICROPHONE},
  {"SRC_PCSPEAKER", MIXERLINE_COMPONENTTYPE_SRC_PCSPEAKER},
  {"SRC_SYNTHESIZER", MIXERLINE_COMPONENTTYPE_SRC_SYNTHESIZER},
  {"SRC_TELEPHONE", MIXERLINE_COMPONENTTYPE_SRC_TELEPHONE},
  {"SRC_UNDEFINED", MIXERLINE_COMPONENTTYPE_SRC_UNDEFINED},
  {"SRC_WAVEOUT", MIXERLINE_COMPONENTTYPE_SRC_WAVEOUT},

  {NULL}

};





ValueNameEntry const ControlTypes [] = {

  {"CUSTOM", MIXERCONTROL_CONTROLTYPE_CUSTOM},
  {"BOOLEANMETER", MIXERCONTROL_CONTROLTYPE_BOOLEANMETER},
  {"SIGNEDMETER", MIXERCONTROL_CONTROLTYPE_SIGNEDMETER},
  {"PEAKMETER", MIXERCONTROL_CONTROLTYPE_PEAKMETER},
  {"UNSIGNEDMETER", MIXERCONTROL_CONTROLTYPE_UNSIGNEDMETER},
  {"BOOLEAN", MIXERCONTROL_CONTROLTYPE_BOOLEAN},
  {"ONOFF", MIXERCONTROL_CONTROLTYPE_ONOFF},
  {"MUTE", MIXERCONTROL_CONTROLTYPE_MUTE},
  {"MONO", MIXERCONTROL_CONTROLTYPE_MONO},
  {"LOUDNESS", MIXERCONTROL_CONTROLTYPE_LOUDNESS},
  {"STEREOENH", MIXERCONTROL_CONTROLTYPE_STEREOENH},
  {"BUTTON", MIXERCONTROL_CONTROLTYPE_BUTTON},
  {"DECIBELS", MIXERCONTROL_CONTROLTYPE_DECIBELS},
  {"SIGNED", MIXERCONTROL_CONTROLTYPE_SIGNED},
  {"UNSIGNED", MIXERCONTROL_CONTROLTYPE_UNSIGNED},
  {"PERCENT", MIXERCONTROL_CONTROLTYPE_PERCENT},
  {"SLIDER", MIXERCONTROL_CONTROLTYPE_SLIDER},
  {"PAN", MIXERCONTROL_CONTROLTYPE_PAN},
  {"QSOUNDPAN", MIXERCONTROL_CONTROLTYPE_QSOUNDPAN},
  {"FADER", MIXERCONTROL_CONTROLTYPE_FADER},
  {"VOLUME", MIXERCONTROL_CONTROLTYPE_VOLUME},
  {"BASS", MIXERCONTROL_CONTROLTYPE_BASS},
  {"TREBLE", MIXERCONTROL_CONTROLTYPE_TREBLE},
  {"EQUALIZER", MIXERCONTROL_CONTROLTYPE_EQUALIZER},
  {"SINGLESELECT", MIXERCONTROL_CONTROLTYPE_SINGLESELECT},
  {"MUX", MIXERCONTROL_CONTROLTYPE_MUX},
  {"MULTIPLESELECT", MIXERCONTROL_CONTROLTYPE_MULTIPLESELECT},
  {"MIXER", MIXERCONTROL_CONTROLTYPE_MIXER},
  {"MICROTIME", MIXERCONTROL_CONTROLTYPE_MICROTIME},
  {"MILLITIME", MIXERCONTROL_CONTROLTYPE_MILLITIME},

  {NULL}

};





ValueNameEntry const ControlStatus [] = {

  {"Disabled", MIXERCONTROL_CONTROLF_DISABLED},
  {"Multiple", MIXERCONTROL_CONTROLF_MULTIPLE},
  {"Uniform", MIXERCONTROL_CONTROLF_UNIFORM},

  {NULL}

};





ValueNameEntry const TargetTypes [] = {

  {"AUX", MIXERLINE_TARGETTYPE_AUX},
  {"MIDIIN", MIXERLINE_TARGETTYPE_MIDIIN},
  {"MIDIOUT", MIXERLINE_TARGETTYPE_MIDIOUT},
  {"WAVEIN", MIXERLINE_TARGETTYPE_WAVEIN},
  {"WAVEOUT", MIXERLINE_TARGETTYPE_WAVEOUT},
  {"UNDEFINED", MIXERLINE_TARGETTYPE_UNDEFINED},

  {NULL}

};





// #########################################################################
//
// Variables
// 
// #########################################################################





HANDLE StdOut = 0;
HANDLE StdErr = 0;





// #########################################################################
//
// Helper functions
// ================
//
// #########################################################################





bool SetConsoleHandles (void) {

  bool Success = false;

  STARTUPINFO SI;

  GetStartupInfo (&SI);

  if (SI.dwFlags & STARTF_USESTDHANDLES) {

    StdOut = SI.hStdOutput;
    StdErr = SI.hStdError;

  } else {

    StdOut = GetStdHandle (STD_OUTPUT_HANDLE);
    StdErr = GetStdHandle (STD_ERROR_HANDLE);

  }

  Success = (StdOut != INVALID_HANDLE_VALUE && StdErr != INVALID_HANDLE_VALUE);

  if (!Success) {

    StdOut = StdErr = 0;

  }

  return Success;

}





void vConsoleOutput (HANDLE Handle, PCSTR Format, va_list Args) {

  char Str [1024];

  UINT const Len = wvsprintf (Str, Format, Args);

  DWORD Written;

  WriteFile (Handle, Str, Len, &Written, 0);
  
}





void _cdecl PrintNoCr (PCSTR Format, ...) {

  va_list Args;

  va_start (Args, Format);
  vConsoleOutput (StdOut, Format, Args);
  va_end (Args);
  
}





void _cdecl Print (PCSTR Format, ...) {

  va_list Args;

  va_start (Args, Format);
  vConsoleOutput (StdOut, Format, Args);
  va_end (Args);
  
  vConsoleOutput (StdOut, "\r\n", 0);

}





void _cdecl Error (PCSTR Format, ...) {

  va_list Args;

  va_start (Args, Format);
  vConsoleOutput (StdErr, Format, Args);
  va_end (Args);
  
  vConsoleOutput (StdErr, "\r\n", 0);

}





void * _cdecl malloc (size_t Size) {

  return (void *) (LocalAlloc (LPTR, Size));

}





void _cdecl free (void *Block) {

  LocalFree (HLOCAL (Block));

}





void * _cdecl operator new (size_t Size) {

  return malloc (Size);

}






void _cdecl operator delete (void *Block) {

  free (Block);

}





String MmSysErr (MMRESULT Res) {

  char Str [MAXERRORLENGTH];

  MMRESULT R = waveInGetErrorText (Res, Str, sizeof (Str));

  if (R != MMSYSERR_NOERROR) {

    wsprintf (Str, "error %u, no description available", Res);

  }

  return Str;

}





// #########################################################################
//
// Mixer functions
// ===============
//
// #########################################################################





String GetValueNames (DWORD Value, ValueNameEntry const *Table, BOOL Equal) {

  char Names [256];

  *Names = '\0';

  DWORD Mask = 0;

  while (Table->Name) {

    bool Found = false;

    if (Equal) {
    
      Found = (Table->Value == Value);
      
    } else {
    
      Mask |= Table->Value;

      Found = ((Table->Value & Value) != 0);

    }

    if (Found) {

      if (*Names) {

        strcat (Names, ",");

      }

      strcat (Names, Table->Name);

      if (Equal) {

        break;

      }

    }

    Table++;

  }

  if (Equal && !*Names) {

    wsprintf (Names, "Unknown (%u, 0x%X)", Value, Value);

  } else if (!Equal && (Value & ~Mask)) {

    wsprintf (Names + strlen (Names), "Unknown (0x%X)", Value & ~Mask);

  }

  return Names;

}





PCSTR GetTabStr (UINT TabSize) {

  static char const Tabs [] = "                                       ";

  return Tabs + sizeof (Tabs) - 1 - TabSize;

}





bool GetLineInfo (UINT DevN, MIXERLINE &Line, DWORD Type) {

  Line.cbStruct = sizeof (Line);

  MMRESULT const Res = mixerGetLineInfo (HMIXEROBJ (DevN), &Line, Type);

  if (Res != MMSYSERR_NOERROR) {

    Error ("Cannot get line info for %u (%s)", Type, MmSysErr (Res));

    return false;

  }

  return true;

}





void PrintCaps (MIXERCAPS const &Caps) {

  Print (
    "wMid:%u, wPid:%u, Ver:%u.%u",
    Caps.wMid, Caps.wPid,
    HIBYTE (Caps.vDriverVersion), LOBYTE (Caps.vDriverVersion)
  );

  Print ("Name:%s", Caps.szPname);

  Print ("Support:%X, DestLines:%u", Caps.fdwSupport, Caps.cDestinations);

}





void PrintControlInfo (UINT DevN, MIXERCONTROL const &Ctrl, UINT Tab) {

  Unused (DevN);

  Print ("");

  Print (
    "%sCtrl:%s (%s), Type:%s, ID:%X",
    GetTabStr (Tab),
    Ctrl.szName, Ctrl.szShortName,
    PCSTR (GetValueNames (Ctrl.dwControlType, ControlTypes, TRUE)),
    Ctrl.dwControlID
  );

  DWORD const Units = Ctrl.dwControlType & MIXERCONTROL_CT_UNITS_MASK;

  if (Units != MIXERCONTROL_CT_UNITS_CUSTOM) {

    if (Units == MIXERCONTROL_CT_UNITS_UNSIGNED) {

      Print (
        "%sBounds:%u..%u",
        GetTabStr (Tab),
        Ctrl.Bounds.dwMinimum, Ctrl.Bounds.dwMaximum
      );

    } else {

      Print (
        "%sBounds:%d..%d",
        GetTabStr (Tab),
        Ctrl.Bounds.lMinimum, Ctrl.Bounds.lMaximum
      );

    }

  }

  Print (
    "%sItems:%u, Status:%s",
    GetTabStr (Tab),
    Ctrl.cMultipleItems,
    PCSTR (GetValueNames (Ctrl.fdwControl, ControlStatus, FALSE))
  );

}





void PrintControlList (UINT DevN, MIXERLINE const &Line, UINT Tab) {

  MIXERLINECONTROLS CtrlInfo;

  CtrlInfo.cbStruct = sizeof (CtrlInfo);

  CtrlInfo.dwLineID = Line.dwLineID;
  CtrlInfo.cControls = Line.cControls;

  MIXERCONTROL Ctrls [20];

  CtrlInfo.pamxctrl = Ctrls;
  CtrlInfo.cbmxctrl = sizeof (*Ctrls);

  MMRESULT Res = mixerGetLineControls (
    HMIXEROBJ (DevN), &CtrlInfo, MIXER_GETLINECONTROLSF_ALL
  );

  if (Res != MMSYSERR_NOERROR) {

    Error ("Cannot get control list (%s)", MmSysErr (Res));

    return;

  }

  for (MIXERCONTROL *Ctrl = Ctrls; Ctrl < Ctrls + CtrlInfo.cControls; Ctrl++) {

    PrintControlInfo (DevN, *Ctrl, Tab + 2);

  }

}





void PrintLineInfo (UINT DevN, MIXERLINE const &Line, UINT Tab) {

  Print ("");

  Print (
    "%sLine:%s (%s), Type:%s, Id:%X",
    GetTabStr (Tab),
    Line.szName, Line.szShortName,
    PCSTR (GetValueNames (Line.dwComponentType, CompTypes, TRUE)),
    Line.dwLineID
  );

  Print (
    "%sDest:%u, Src:%d, Channs:%u, Conns:%u, Ctrls:%u, Status:%s",
    GetTabStr (Tab),
    Line.dwDestination, Line.dwSource,
    Line.cChannels, Line.cConnections, Line.cControls,
    PCSTR (GetValueNames (Line.fdwLine, LineStatusFlags, FALSE))
  );

  if (Line.Target.dwType != MIXERLINE_TARGETTYPE_UNDEFINED) {

    Print (
      "%sTarget:%s, DevId:%d (%s)",
      GetTabStr (Tab),
      PCSTR (GetValueNames (Line.Target.dwType, TargetTypes, TRUE)),
      Line.Target.dwDeviceID, Line.Target.szPname
    );

  }

  if (Line.cControls) {

    PrintControlList (DevN, Line, Tab);

  }

}





void PrintDestLines (UINT DevN, DWORD NumLines) {

  if (NumLines) {

    Print ("Line info:");

  }

  for (DWORD N = 0; N < NumLines; N++) {

    MIXERLINE Line;

    Line.dwDestination = N;

    if (GetLineInfo (DevN, Line, MIXER_GETLINEINFOF_DESTINATION)) {

      PrintLineInfo (DevN, Line, 2);

      if (!(Line.fdwLine & MIXERLINE_LINEF_SOURCE) && Line.cConnections) {

        Print ("");
        Print ("  Source lines:");

        for (DWORD NCon = 0; NCon < Line.cConnections; NCon++) {

          MIXERLINE SrcLine;

          SrcLine.dwDestination = Line.dwDestination;
          SrcLine.dwSource = NCon;

          if (GetLineInfo (DevN, SrcLine, MIXER_GETLINEINFOF_SOURCE)) {

            PrintLineInfo (DevN, SrcLine, 4);

          }

        }

      }

    }

  }

}





// #########################################################################
//
// Main function
// =============
//
// #########################################################################





DWORD WINAPI WinEntry (void *Arg) {

  Unused (Arg);

  UINT Result = 0;


  do {

    if (!SetConsoleHandles ()) {

      Result = 99;

      break;

    }

    int const NumDevs = mixerGetNumDevs ();

    PCSTR p = GetCommandLine ();

    char Delim = *p;
    
    if (Delim == '"') {
    
      ++p;
      
    } else {
    
      Delim = ' ';

    }

    PCSTR const p2 = strchr (p, Delim);

    if (p2) {

      p = p2 + 1;

    } else {

      p += strlen (p);

    }

    while (*p == ' ') ++p;

    if (!*p) {

      Print (
        "Command line: <DevNum>\n"
        "  DevNum = 0..<number of mixer devices minus one>\n"
        "  There are %u mixer devices in your system",
        NumDevs
      );

      Result = 1;

      break;

    }

    int const DevNum = StrToInt (p);

    if (DevNum >= NumDevs) {

      Error ("No mixer device %u", DevNum);

      Result = 2;

      break;

    }

    Print ("");

    MMRESULT Res;

    Print ("Info of mixer %u:", DevNum);

    MIXERCAPS Caps;

    Res = mixerGetDevCaps (DevNum, &Caps, sizeof (Caps));

    if (Res != MMSYSERR_NOERROR) {

      Error ("Error getting mixer caps (%s)", MmSysErr (Res));

      break;

    }

    PrintCaps (Caps);


    Print ("");
    PrintDestLines (DevNum, Caps.cDestinations);


  } while (0);

  ExitProcess (Result);

  return 0;

}
